# Standard Library
import json
from unittest.mock import patch

# Third Party
import requests
from django_webtest import WebTest

# Django
from django.urls import reverse

# AA Forum
from aa_forum.models import Board, Category, Message, Topic
from aa_forum.tests.utils import (
    create_fake_message,
    create_fake_messages,
    create_fake_user,
)


class TestForumUI(WebTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user_1001 = create_fake_user(
            1001, "Bruce Wayne", permissions=["aa_forum.basic_access"]
        )
        cls.user_1002 = create_fake_user(
            1002, "Peter Parker", permissions=["aa_forum.basic_access"]
        )
        cls.user_1003 = create_fake_user(1003, "Lex Luthor", permissions=[])
        cls.category = Category.objects.create(name="Science")
        cls.board = Board.objects.create(name="Physics", category=cls.category)
        cls.board_with_webhook = Board.objects.create(
            name="Chemistry",
            category=cls.category,
            discord_webhook="https://discord.com/webhook/",
        )
        cls.board_with_webhook_for_replies = Board.objects.create(
            name="Biology",
            category=cls.category,
            discord_webhook="https://discord.com/webhook/",
            use_webhook_for_replies=True,
        )

    def test_should_show_forum_index(self):
        # given
        self.app.set_user(self.user_1001)
        # when
        page = self.app.get(reverse("aa_forum:forum_index"))
        # then
        self.assertTemplateUsed(page, "aa_forum/view/forum/index.html")

    def test_should_not_show_forum_index(self):
        # given
        self.app.set_user(self.user_1003)
        # when
        page = self.app.get(reverse("aa_forum:forum_index"))
        # then
        self.assertRedirects(page, "/account/login/?next=/forum/")

    def test_should_create_new_topic(self):
        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board.get_absolute_url())
        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")

        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Recent Discoveries"
        form["message"] = "Energy of the Higgs boson"
        page = form.submit().follow()

        # then
        self.assertEqual(self.board.topics.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_topic = Topic.objects.last()
        self.assertEqual(new_topic.subject, "Recent Discoveries")
        new_message = Message.objects.last()
        self.assertEqual(new_message.message, "Energy of the Higgs boson")

    @patch("requests.post")
    def test_should_post_to_discord_webhook_on_create_new_topic(self, mock_post):
        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board_with_webhook.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")

        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Recent Discoveries"
        form["message"] = "Energy of the Higgs boson"
        page = form.submit().follow()

        # then
        self.assertEqual(self.board_with_webhook.topics.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_topic = Topic.objects.last()
        self.assertEqual(new_topic.subject, "Recent Discoveries")
        new_message = Message.objects.last()
        self.assertEqual(new_message.message, "Energy of the Higgs boson")

        info = {"test1": "value1", "test2": "value2"}
        requests.post(self.board_with_webhook.discord_webhook, data=json.dumps(info))
        mock_post.assert_called_with(
            self.board_with_webhook.discord_webhook, data=json.dumps(info)
        )

    @patch("requests.post")
    def test_should_post_to_discord_webhook_with_image_on_create_new_topic(
        self, mock_post
    ):
        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board_with_webhook.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")

        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Recent Discoveries"
        form[
            "message"
        ] = "Energy of the Higgs boson <img src='/images/images/038/929/227/large/marc-bell-2a.jpg'>"
        page = form.submit().follow()

        # then
        self.assertEqual(self.board_with_webhook.topics.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_topic = Topic.objects.last()
        self.assertEqual(new_topic.subject, "Recent Discoveries")
        new_message = Message.objects.last()
        self.assertEqual(
            new_message.message,
            "Energy of the Higgs boson <img src='/images/images/038/929/227/large/marc-bell-2a.jpg'>",
        )

        info = {"test1": "value1", "test2": "value2"}
        requests.post(self.board_with_webhook.discord_webhook, data=json.dumps(info))
        mock_post.assert_called_with(
            self.board_with_webhook.discord_webhook, data=json.dumps(info)
        )

    @patch("requests.post")
    def test_should_post_to_discord_webhook_with_image_with_full_url_on_create_new_topic(
        self, mock_post
    ):
        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board_with_webhook.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")

        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Recent Discoveries"
        form[
            "message"
        ] = "Energy of the Higgs boson <img src='https://cdnb.artstation.com/p/assets/images/images/038/929/227/large/marc-bell-2a.jpg'>"
        page = form.submit().follow()

        # then
        self.assertEqual(self.board_with_webhook.topics.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_topic = Topic.objects.last()
        self.assertEqual(new_topic.subject, "Recent Discoveries")
        new_message = Message.objects.last()
        self.assertEqual(
            new_message.message,
            "Energy of the Higgs boson <img src='https://cdnb.artstation.com/p/assets/images/images/038/929/227/large/marc-bell-2a.jpg'>",
        )

        info = {"test1": "value1", "test2": "value2"}
        requests.post(self.board_with_webhook.discord_webhook, data=json.dumps(info))
        mock_post.assert_called_with(
            self.board_with_webhook.discord_webhook, data=json.dumps(info)
        )

    def test_should_cancel_new_topic(self):
        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board.get_absolute_url())
        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")
        # then
        page = page.click(linkid="aa-forum-btn-cancel")
        # then
        self.assertEqual(self.board.topics.count(), 0)
        self.assertTemplateUsed(page, "aa_forum/view/forum/board.html")

    def test_should_create_reply_in_topic(self):
        # given
        topic = Topic.objects.create(
            subject="Mysteries", board=self.board_with_webhook_for_replies
        )
        create_fake_messages(topic=topic, amount=5)
        self.app.set_user(self.user_1001)
        page = self.app.get(topic.get_absolute_url())

        # when
        form = page.forms["aa-forum-form-message-reply"]
        form["message"] = "What is dark matter?"
        page = form.submit().follow().follow()

        # then
        self.assertEqual(topic.messages.count(), 6)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_message = Message.objects.last()
        self.assertEqual(new_message.message, "What is dark matter?")

    def test_should_post_to_webhook_on_create_reply_in_topic(self):
        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_messages(topic=topic, amount=5)
        self.app.set_user(self.user_1001)
        page = self.app.get(topic.get_absolute_url())
        # when
        form = page.forms["aa-forum-form-message-reply"]
        form["message"] = "What is dark matter?"
        page = form.submit().follow().follow()
        # then
        self.assertEqual(topic.messages.count(), 6)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_message = Message.objects.last()
        self.assertEqual(new_message.message, "What is dark matter?")

    def test_should_update_own_message(self):
        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        own_message = create_fake_message(topic=topic, user=self.user_1001)
        self.app.set_user(self.user_1001)
        page = self.app.get(topic.get_absolute_url())
        # when
        page = page.click(linkid=f"aa-forum-btn-modify-message-{own_message.pk}")
        form = page.forms["aa-forum-form-message-modify"]
        form["message"] = "What is dark matter?"
        page = form.submit().follow().follow()
        # then
        self.assertEqual(topic.messages.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        own_message.refresh_from_db()
        self.assertEqual(own_message.message, "What is dark matter?")

    def test_should_not_be_able_to_edit_messages_from_others(self):
        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        alien_message = create_fake_message(topic=topic, user=self.user_1002)
        self.app.set_user(self.user_1001)
        # when
        page = self.app.get(topic.get_absolute_url())
        # then
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        self.assertNotContains(page, f"aa-forum-btn-modify-message-{alien_message.pk}")

    def test_should_find_message_by_key_word(self):
        # given
        topic_1 = Topic.objects.create(subject="Topic 1", board=self.board)
        create_fake_messages(topic=topic_1, amount=5)
        topic_2 = Topic.objects.create(subject="Topic 2", board=self.board)
        create_fake_messages(topic=topic_2, amount=5)
        message = Message.objects.create(
            topic=topic_1, user_created=self.user_1001, message="xyz dummy123 abc"
        )
        self.app.set_user(self.user_1001)
        page = self.app.get(reverse("aa_forum:forum_index"))
        # when
        form = page.forms["aa-forum-form-search-menu"]
        form["q"] = "dummy123"
        res = form.submit()
        # then
        self.assertTemplateUsed(res, "aa_forum/view/search/results.html")
        self.assertContains(
            res,
            reverse(
                "aa_forum:forum_message",
                args=[
                    message.topic.board.category.slug,
                    message.topic.board.slug,
                    message.topic.slug,
                    message.pk,
                ],
            ),
        )


class TestAdminUI(WebTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = create_fake_user(
            1001,
            "Bruce Wayne",
            permissions=["aa_forum.basic_access", "aa_forum.manage_forum"],
        )

    def test_should_create_category(self):
        # given
        self.app.set_user(self.user)
        page = self.app.get(reverse("aa_forum:admin_index"))
        # when
        form = page.forms["aa-forum-form-admin-new-category"]
        form["new-category-name"] = "Category"
        page = form.submit().follow()
        # then
        self.assertTemplateUsed(page, "aa_forum/view/administration/index.html")
        new_category = Category.objects.last()
        self.assertEqual(new_category.name, "Category")

    def test_should_edit_category(self):
        # given
        category = Category.objects.create(name="Category")
        self.app.set_user(self.user)
        page = self.app.get(reverse("aa_forum:admin_index"))
        # when
        form = page.forms[f"aa-forum-form-admin-edit-category-{category.pk}"]
        form[f"edit-category-{category.pk}-name"] = "Dummy"
        page = form.submit().follow()
        # then
        self.assertTemplateUsed(page, "aa_forum/view/administration/index.html")
        category.refresh_from_db()
        self.assertEqual(category.name, "Dummy")

    def test_should_add_board_to_category(self):
        # given
        category = Category.objects.create(name="Category")
        self.app.set_user(self.user)
        page = self.app.get(reverse("aa_forum:admin_index"))
        # when
        form = page.forms[f"aa-forum-form-admin-add-board-{category.id}"]
        form[f"new-board-in-category-{category.id}-name"] = "Board"
        page = form.submit().follow()
        # then
        self.assertTemplateUsed(page, "aa_forum/view/administration/index.html")
        new_board = category.boards.last()
        self.assertEqual(new_board.name, "Board")

    def test_should_edit_board(self):
        # given
        category = Category.objects.create(name="Category")
        board = Board.objects.create(name="Board", category=category)
        self.app.set_user(self.user)
        page = self.app.get(reverse("aa_forum:admin_index"))
        # when
        form = page.forms[f"aa-forum-form-edit-board-{board.pk}"]
        form[f"edit-board-{board.pk}-name"] = "Dummy"
        page = form.submit().follow()
        # then
        self.assertTemplateUsed(page, "aa_forum/view/administration/index.html")
        board.refresh_from_db()
        self.assertEqual(board.name, "Dummy")



"""
Forum related views
"""

# Standard Library
from typing import Optional

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.models import Group
from django.core.handlers.wsgi import WSGIRequest
from django.core.paginator import Paginator
from django.db import transaction
from django.db.models import Count, Exists, OuterRef, Prefetch, Q
from django.http import HttpResponse, HttpResponseNotFound, HttpResponseRedirect
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Forum
from aa_forum import __title__
from aa_forum.constants import SETTING_MESSAGESPERPAGE, SETTING_TOPICSPERPAGE
from aa_forum.forms import EditMessageForm, EditTopicForm, NewTopicForm
from aa_forum.helper.webhook import send_message_to_discord_webhook
from aa_forum.models import Board, Category, LastMessageSeen, Message, Setting, Topic

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("aa_forum.basic_access")
def index(request: WSGIRequest) -> HttpResponse:
    """
    Forum index view
    :param request:
    :type request:
    :return:
    :rtype:
    """

    has_read_all_messages = LastMessageSeen.objects.filter(
        topic=OuterRef("pk"),
        user=request.user,
        message_time__gte=OuterRef("last_message__time_posted"),
    )
    unread_topic_pks = Topic.objects.filter(~Exists(has_read_all_messages)).values_list(
        "pk", flat=True
    )

    boards = (
        Board.objects.select_related(
            "category",
            "last_message",
            "last_message__topic",
            "last_message__user_created__profile__main_character",
            "first_message",
        )
        .prefetch_related(
            Prefetch(
                "child_boards",
                queryset=Board.objects.select_related(
                    "category",
                    "parent_board",
                    "last_message",
                    "last_message__user_created",
                    "last_message__user_created__profile__main_character",
                    "first_message",
                    "first_message__user_created",
                    "first_message__user_created__profile__main_character",
                )
                .annotate(
                    num_posts=Count("topics__messages", distinct=True),
                    num_topics=Count("topics", distinct=True),
                    num_unread=Count(
                        "topics", filter=Q(topics__in=unread_topic_pks), distinct=True
                    ),
                )
                .user_has_access(request.user)
                .order_by("order", "id"),
            )
        )
        .prefetch_related(Prefetch("groups", queryset=Group.objects.order_by("name")))
        .prefetch_related("topics")
        .user_has_access(request.user)
        .filter(parent_board__isnull=True)
        .annotate(
            num_posts=Count("topics__messages", distinct=True),
            num_topics=Count("topics", distinct=True),
            num_unread=Count(
                "topics", filter=Q(topics__in=unread_topic_pks), distinct=True
            ),
        )
        .order_by("category__order", "category__id", "order", "id")
    )

    categories_map = dict()

    for board in boards:
        category = board.category

        if category.pk not in categories_map:
            categories_map[category.pk] = {
                "id": category.id,
                "name": category.name,
                "boards_sorted": list(),
                "order": category.order,
            }

        categories_map[category.pk]["boards_sorted"].append(board)

    categories = sorted(categories_map.values(), key=lambda k: k["order"])
    context = {"categories": categories}

    logger.info(f"{request.user} called forum index")

    return render(request, "aa_forum/view/forum/index.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def board(
    request: WSGIRequest, category_slug: str, board_slug: str, page_number: int = None
) -> HttpResponse:
    """
    Forum board view
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param page_number:
    :type page_number:
    :return:
    :rtype:
    """

    has_read_all_messages = LastMessageSeen.objects.filter(
        topic=OuterRef("pk"),
        user=request.user,
        message_time__gte=OuterRef("last_message__time_posted"),
    )
    unread_topic_pks = Topic.objects.filter(~Exists(has_read_all_messages)).values_list(
        "pk", flat=True
    )

    try:
        board = (
            Board.objects.select_related("category")
            .select_related("parent_board")
            .prefetch_related(
                Prefetch(
                    "child_boards",
                    queryset=Board.objects.select_related(
                        "category",
                        "parent_board",
                        "last_message",
                        "last_message__user_created",
                        "last_message__user_created__profile__main_character",
                        "first_message",
                        "first_message__user_created",
                        "first_message__user_created__profile__main_character",
                    )
                    .prefetch_related(
                        Prefetch("groups", queryset=Group.objects.order_by("name"))
                    )
                    .prefetch_related("topics")
                    .annotate(
                        num_posts=Count("topics__messages", distinct=True),
                        num_topics=Count("topics", distinct=True),
                        num_unread=Count(
                            "topics",
                            filter=Q(topics__in=unread_topic_pks),
                            distinct=True,
                        ),
                    )
                    .user_has_access(request.user)
                    .order_by("order", "id"),
                )
            )
            .prefetch_related(
                Prefetch(
                    "topics",
                    queryset=Topic.objects.select_related(
                        "last_message",
                        "last_message__user_created",
                        "last_message__user_created__profile__main_character",
                        "first_message",
                        "first_message__user_created",
                        "first_message__user_created__profile__main_character",
                    )
                    .annotate(num_posts=Count("messages", distinct=True))
                    .annotate(has_unread_messages=~Exists(has_read_all_messages))
                    .order_by("-is_sticky", "-last_message__time_posted", "-id"),
                    to_attr="topics_sorted",
                )
            )
            .filter(category__slug=category_slug, slug=board_slug)
            .user_has_access(request.user)
            .get()
        )
    except Board.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The board you were trying to visit does "
                    "either not exist, or you don't have access to it ...</p>"
                )
            ),
        )

        logger.info(
            f"{request.user} called board without having access to it. "
            f"Redirecting to forum index"
        )

        return redirect("aa_forum:forum_index")

    paginator = Paginator(
        board.topics_sorted,
        int(Setting.objects.get_setting(setting_key=SETTING_TOPICSPERPAGE)),
    )
    page_obj = paginator.get_page(page_number)
    context = {"board": board, "page_obj": page_obj}

    logger.info(f'{request.user} called board "{board.name}"')

    return render(request, "aa_forum/view/forum/board.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def board_new_topic(
    request: WSGIRequest, category_slug: str, board_slug: str
) -> HttpResponse:
    """
    Begin a new topic
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :return:
    :rtype:
    """

    try:
        Category.objects.get(slug__exact=category_slug)
    except Category.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The category you were trying to post in does "
                    "not exist ...</p>"
                )
            ),
        )

        logger.info(f"{request.user} tried to open a non existing category")

        return redirect("aa_forum:forum_index")

    try:
        board = (
            Board.objects.select_related("category")
            .user_has_access(request.user)
            .filter(category__slug=category_slug, slug=board_slug)
            .get()
        )
    except Board.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The board you were trying to post in does "
                    "either not exist, or you don't have access to it ...</p>"
                )
            ),
        )

        logger.info(
            f"{request.user} tried to create a topic in a board they have no access "
            f"to. Redirecting to forum index"
        )

        return redirect("aa_forum:forum_index")

    # If this is a POST request we need to process the form data ...
    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = NewTopicForm(request.POST)

        # Check whether it's valid:
        if form.is_valid():
            with transaction.atomic():
                # Check if a topic with the same subject already exists in this board
                existing_topic = Topic.objects.filter(
                    board=board, subject__iexact=form.cleaned_data["subject"]
                )

                if existing_topic.exists():
                    existing_topic = existing_topic.get()

                    existing_topic_url = reverse(
                        "aa_forum:forum_topic",
                        kwargs={
                            "category_slug": existing_topic.board.category.slug,
                            "board_slug": existing_topic.board.slug,
                            "topic_slug": existing_topic.slug,
                        },
                    )

                    messages.warning(
                        request,
                        mark_safe(
                            _(
                                "<h4>Warning!</h4>"
                                f"<p>There is already a topic with the exact same "
                                f"subject in this board.</p><p>See here: "
                                f'<a href="{existing_topic_url}">'
                                f"{existing_topic.subject}</a>"
                                f"</p>"
                            )
                        ),
                    )

                    return render(
                        request,
                        "aa_forum/view/forum/new-topic.html",
                        {"board": board, "form": form},
                    )

                topic = Topic(board=board, subject=form.cleaned_data["subject"])
                topic.save()

                message = Message(
                    topic=topic,
                    user_created=request.user,
                    message=form.cleaned_data["message"],
                )
                message.save()

            logger.info(
                f'{request.user} started a new topic "{topic.subject}" '
                f'in board "{board.name}"'
            )

            # Send to webhook if one is configured
            if board.discord_webhook is not None:
                send_message_to_discord_webhook(
                    board=board,
                    topic=topic,
                    message=message,
                    headline=f'**New topic has been started in board "{board.name}"**',
                )

            return redirect(
                "aa_forum:forum_topic",
                category_slug=board.category.slug,
                board_slug=board.slug,
                topic_slug=topic.slug,
            )
    # If not, we'll create a blank form
    else:
        form = NewTopicForm()

    context = {"board": board, "form": form}

    logger.info(f'{request.user} is starting a new topic in board "{board.name}"')

    return render(request, "aa_forum/view/forum/new-topic.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def topic(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
    page_number: int = None,
) -> HttpResponse:
    """
    View a topic
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :param page_number:
    :type page_number:
    :return:
    :rtype:
    """

    topic = _topic_from_slugs(
        request=request,
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
    )

    if not topic:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The topic you were trying to view does not "
                    "exist or you do not have access to it.</p>"
                )
            ),
        )

        logger.info(
            f"{request.user} called a non existent topic. Redirecting to forum index"
        )

        return redirect("aa_forum:forum_index")

    # Determine if the current user can modify the topics subject
    can_modify_subject = False
    if request.user == topic.first_message.user_created or request.user.has_perm(
        "aa_forum.manage_forum"
    ):
        can_modify_subject = True

    # Set this topic as "read by" by the current user
    paginator = Paginator(
        topic.messages_sorted,
        int(Setting.objects.get_setting(setting_key=SETTING_MESSAGESPERPAGE)),
    )
    page_obj = paginator.get_page(page_number)

    try:
        last_message_on_page = page_obj.object_list[-1]
    except IndexError:
        pass
    else:
        try:
            last_message_seen = LastMessageSeen.objects.get(
                topic=topic, user=request.user
            )
        except LastMessageSeen.DoesNotExist:
            last_message_seen = None

        if (
            not last_message_seen
            or last_message_seen.message_time < last_message_on_page.time_posted
        ):
            LastMessageSeen.objects.update_or_create(
                topic=topic,
                user=request.user,
                defaults={"message_time": last_message_on_page.time_posted},
            )

    context = {
        "topic": topic,
        "can_modify_subject": can_modify_subject,
        "page_obj": page_obj,
        "reply_form": EditMessageForm(),
    }

    logger.info(f'{request.user} called topic "{topic.subject}"')

    return render(request, "aa_forum/view/forum/topic.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def topic_modify(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
) -> HttpResponse:
    """
    Modify the topics subject
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    """

    topic_to_modify = _topic_from_slugs(
        request=request,
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
    )

    if not topic_to_modify:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The topic you were trying to modify does not "
                    "exist or you do not have access to it.</p>"
                )
            ),
        )

        logger.info(
            f"{request.user} called a non existent topic. Redirecting to forum index"
        )

        return redirect("aa_forum:forum_index")

    # Check if the user actually has the right to edit this message
    if (
        topic_to_modify.first_message.user_created != request.user
        and not request.user.has_perm("aa_forum.manage_forum")
    ):
        messages.error(
            request,
            mark_safe(
                _("<h4>Error!</h4><p>You are not allowed to modify this topic!</p>")
            ),
        )

        logger.info(
            f'{request.user} tried to modify topic "{topic_to_modify.subject}" '
            f"without the proper permissions. Redirecting to forum index"
        )

        return redirect(
            "aa_forum:forum_topic",
            category_slug=category_slug,
            board_slug=board_slug,
            topic_slug=topic_slug,
        )

    # We are in the clear, let's see what we've got
    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditTopicForm(request.POST)

        # Check whether it's valid:
        if form.is_valid():
            topic_to_modify.subject = form.cleaned_data["subject"]
            topic_to_modify.save()

            messages.success(
                request,
                mark_safe(
                    _("<h4>Success!</h4><p>The topic subject has been updated.</p>")
                ),
            )

            logger.info(f'{request.user} modified topic "{topic_to_modify.subject}"')

            return redirect(
                "aa_forum:forum_topic",
                category_slug=category_slug,
                board_slug=board_slug,
                topic_slug=topic_slug,
            )
    # If not, we'll fill the form with the information from the message object
    else:
        form = EditTopicForm(instance=topic_to_modify)

    context = {"form": form, "topic": topic_to_modify}

    logger.info(f'{request.user} modifying "{topic_to_modify.subject}"')

    return render(request, "aa_forum/view/forum/modify-topic.html", context)


def _topic_from_slugs(
    request: WSGIRequest, category_slug: str, board_slug: str, topic_slug: str
) -> Optional[Topic]:
    """
    Fetch topic from given slug params.
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :return:
    :rtype:
    """

    topic = Topic.objects.get_from_slugs(
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
        user=request.user,
    )

    return topic


@login_required
@permission_required("aa_forum.basic_access")
def topic_first_unread_message(
    request: WSGIRequest, category_slug: str, board_slug: str, topic_slug: str
) -> HttpResponse:
    """
    Redirect to first unread message of a topic.
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :return:
    :rtype:
    """

    topic = _topic_from_slugs(
        request=request,
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
    )

    if not topic:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The topic you were trying to view does not "
                    "exist or you do not have access to it.</p>"
                )
            ),
        )

        return redirect("aa_forum:forum_index")

    messages_sorted = topic.messages.order_by("time_posted")

    try:
        last_message_seen = LastMessageSeen.objects.filter(
            topic=topic, user=request.user
        ).get()
    except LastMessageSeen.DoesNotExist:
        redirect_message = messages_sorted.first()
    else:
        redirect_message = messages_sorted.filter(
            time_posted__gt=last_message_seen.message_time
        ).first()

        if not redirect_message:
            redirect_message = messages_sorted.last()

    if redirect_message:
        return redirect(redirect_message.get_absolute_url())

    return redirect(topic.get_absolute_url())


@login_required
@permission_required("aa_forum.basic_access")
def topic_show_all_unread(request: WSGIRequest) -> HttpResponse:
    """
    Show all available unread topics
    :param request:
    :type request:
    """

    has_read_all_messages = LastMessageSeen.objects.filter(
        topic=OuterRef("pk"),
        user=request.user,
        message_time__gte=OuterRef("last_message__time_posted"),
    )
    unread_topic_pks = Topic.objects.filter(~Exists(has_read_all_messages)).values_list(
        "pk", flat=True
    )

    boards = (
        Board.objects.select_related(
            "parent_board",
            "category",
            "last_message",
            "last_message__topic",
            "last_message__user_created__profile__main_character",
            "first_message",
        )
        .prefetch_related(
            Prefetch(
                "topics",
                queryset=Topic.objects.select_related(
                    "last_message",
                    "last_message__user_created",
                    "last_message__user_created__profile__main_character",
                    "first_message",
                    "first_message__user_created",
                    "first_message__user_created__profile__main_character",
                )
                .filter(pk__in=unread_topic_pks)
                .annotate(num_posts=Count("messages", distinct=True))
                .annotate(has_unread_messages=~Exists(has_read_all_messages))
                .order_by("-is_sticky", "-last_message__time_posted", "-id"),
            )
        )
        .filter(topics__in=unread_topic_pks)
        .user_has_access(request.user)
        .order_by("category__order", "category__id", "order", "id")
        .all()
    )

    context = {"boards": boards}

    logger.info(f"{request.user} calling unread topics view")

    return render(request, "aa_forum/view/forum/unread-topics.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def topic_reply(
    request: WSGIRequest, category_slug: str, board_slug: str, topic_slug: str
) -> HttpResponse:
    """
    Reply to a post in a topic
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :return:
    :rtype:
    """

    topic = _topic_from_slugs(
        request=request,
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
    )

    if not topic:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The topic you were trying to reply does not "
                    "exist or you do not have access to it.</p>"
                )
            ),
        )

        logger.info(
            f"{request.user} called a non existent topic. Redirecting to forum index"
        )

        return redirect("aa_forum:forum_index")

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditMessageForm(request.POST)

        # Check whether it's valid:
        if form.is_valid():
            new_message = Message(
                topic=topic,
                user_created=request.user,
                message=form.cleaned_data["message"],
            )
            new_message.save()

            # Send to webhook if one is configured
            if (
                topic.board.discord_webhook is not None
                and topic.board.use_webhook_for_replies is not False
            ):
                send_message_to_discord_webhook(
                    board=topic.board,
                    topic=topic,
                    message=new_message,
                    headline=f'**New reply has been posted in topic "{topic.subject}"**',
                )

            logger.info(f"{request.user} replied to topic {topic.subject}")

            return redirect(
                "aa_forum:forum_message",
                category_slug=category_slug,
                board_slug=board_slug,
                topic_slug=topic_slug,
                message_id=new_message.id,
            )

    messages.warning(
        request,
        mark_safe(_("<h4>Warning!</h4><p>Something went wrong, please try again</p>.")),
    )

    return redirect("aa_forum:forum_topic", category_slug, board_slug, topic_slug)


@login_required
@permission_required("aa_forum.manage_forum")
def topic_change_lock_state(
    request: WSGIRequest, topic_id: int
) -> HttpResponseRedirect:
    """
    Change the lock state of the given topic
    :param request:
    :type request:
    :param topic_id:
    :type topic_id:
    :return:
    :rtype:
    """

    try:
        topic = Topic.objects.select_related("board").get(pk=topic_id)
    except Topic.DoesNotExist:
        return HttpResponseNotFound("Could not find topic.")

    if topic.is_locked:
        topic.is_locked = False

        messages.success(
            request,
            mark_safe(_("<h4>Success!</h4><p>Topic has been unlocked/re-opened.</p>")),
        )

        logger.info(f"{request.user} unlocked/re-opened topic {topic.subject}")
    else:
        topic.is_locked = True

        messages.success(
            request,
            mark_safe(_("<h4>Success!</h4><p>Topic has been locked/closed.</p>")),
        )

        logger.info(f'{request.user} locked/closed "{topic.subject}"')

    topic.save(update_fields=["is_locked"])

    return redirect("aa_forum:forum_board", topic.board.category.slug, topic.board.slug)


@login_required
@permission_required("aa_forum.manage_forum")
def topic_change_sticky_state(
    request: WSGIRequest, topic_id: int
) -> HttpResponseRedirect:
    """
    Change the sticky state of the given topic
    :param request:
    :type request:
    :param topic_id:
    :type topic_id:
    :return:
    :rtype:
    """

    try:
        topic = Topic.objects.select_related("board").get(pk=topic_id)
    except Topic.DoesNotExist:
        return HttpResponseNotFound("Could not find topic.")

    if topic.is_sticky:
        topic.is_sticky = False

        messages.success(
            request,
            mark_safe(_('<h4>Success!</h4><p>Topic is no longer "Sticky".</p>')),
        )

        logger.info(
            f'{request.user} changed topic "{topic.subject}" to be no longer sticky'
        )
    else:
        topic.is_sticky = True

        messages.success(
            request,
            mark_safe(_('<h4>Success!</h4><p>Topic is now "Sticky".</p>')),
        )

        logger.info(f'{request.user} changed topic "{topic.subject}" to be sticky')

    topic.save(update_fields=["is_sticky"])

    return redirect(topic.board.get_absolute_url())


@login_required
@permission_required("aa_forum.manage_forum")
def topic_delete(request: WSGIRequest, topic_id: int) -> HttpResponseRedirect:
    """
    Delete a given topic
    :param request:
    :type request:
    :param topic_id:
    :type topic_id:
    :return:
    :rtype:
    """

    try:
        topic = Topic.objects.select_related("board").get(pk=topic_id)
    except Topic.DoesNotExist:
        return HttpResponseNotFound("Could not find topic.")

    board = topic.board
    topic_subject = topic.subject

    topic.delete()

    messages.success(
        request,
        mark_safe(_("<h4>Success!</h4><p>Topic removed.</p>")),
    )

    logger.info(f'{request.user} removed topic "{topic_subject}"')

    return redirect(board.get_absolute_url())


@login_required
@permission_required("aa_forum.basic_access")
def message(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
    message_id: int,
) -> HttpResponseRedirect:
    """
    Get a messages' entry point in a topic, so we end up on the right page with it
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :param message_id:
    :type message_id:
    :return:
    :rtype:
    """

    try:
        message = Message.objects.select_related(
            "topic", "topic__board", "topic__board__category"
        ).get(pk=message_id)
    except Message.DoesNotExist:
        messages.error(
            request,
            mark_safe(_("<h4>Error!</h4><p>The message doesn't exist ...</p>")),
        )

        return redirect("aa_forum:forum_index")

    return redirect(message.get_absolute_url())


@login_required
@permission_required("aa_forum.basic_access")
def message_modify(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
    message_id: int,
) -> HttpResponse:
    """
    Modify a given message
    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :param message_id:
    :type message_id:
    :return:
    :rtype:
    """

    message_to_modify = _message_from_slugs(
        request=request,
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
        message_id=message_id,
    )

    # Check if the message exists.
    # If not, the user doesn't have access to its board, or it doesn't exist.
    # Either way, message can't be edited
    if not message_to_modify:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4><p>The message you were trying to modify does "
                    "either not exist, or you don't have access to it ...</p>"
                )
            ),
        )

        logger.info(
            f"{request.user} trying to change a message in a topic that either does "
            f"not exist or they have no access to"
        )

        return redirect("aa_forum:forum_index")

    # Check if the user actually has the right to edit this message
    if (
        message_to_modify.user_created_id != request.user.id
        and not request.user.has_perm("aa_forum.manage_forum")
    ):
        messages.error(
            request,
            mark_safe(
                _("<h4>Error!</h4><p>You are not allowed to modify this message!</p>")
            ),
        )

        logger.info(
            f"{request.user} tried to modify a message in topic "
            f'"{message_to_modify.topic.subject}" without permission to do so'
        )

        return redirect(message_to_modify.topic.get_absolute_url())

    # We are in the clear, let's see what we've got
    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditMessageForm(request.POST)

        # Check whether it's valid:
        if form.is_valid():
            message_to_modify.user_updated = request.user
            message_to_modify.message = form.cleaned_data["message"]
            message_to_modify.save()

            messages.success(
                request,
                mark_safe(_("<h4>Success!</h4><p>The message has been updated.</p>")),
            )

            logger.info(
                f"{request.user} modified message ID {message_to_modify.pk} "
                f'in topic "{message_to_modify.topic.subject}"'
            )

            return redirect(
                "aa_forum:forum_message",
                category_slug=category_slug,
                board_slug=board_slug,
                topic_slug=topic_slug,
                message_id=message_id,
            )
    # If not, we'll fill the form with the information from the message object
    else:
        form = EditMessageForm(instance=message_to_modify)

    context = {
        "form": form,
        "board": message_to_modify.topic.board,
        "message": message_to_modify,
    }

    logger.info(
        f"{request.user} is modifying message ID {message_to_modify.pk} "
        f'in topic "{message_to_modify.topic.subject}"'
    )

    return render(request, "aa_forum/view/forum/modify-message.html", context)


def _message_from_slugs(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
    message_id: int,
) -> Optional[Message]:
    """

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :param message_id:
    :type message_id:
    :return:
    :rtype:
    """

    message = Message.objects.get_from_slugs(
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
        message_id=message_id,
        user=request.user,
    )

    return message


@login_required
@permission_required("aa_forum.basic_access")
def message_delete(request: WSGIRequest, message_id: int) -> HttpResponseRedirect:
    """
    Delete a message from a topic
    If it is the last message in this topic, the topic will be removed as well
    :param request:
    :type request:
    :param message_id:
    :type message_id:
    :return:
    :rtype:
    """

    try:
        message = (
            Message.objects.select_related(
                "topic", "topic__board", "topic__board__category"
            )
            .user_has_access(request.user)
            .get(pk=message_id)
        )
    except Message.DoesNotExist:
        return HttpResponseNotFound("Message not found.")

    topic = message.topic
    topic_subject = topic.subject

    # Safety check to make sure the user is allowed to delete this message
    if message.user_created_id != request.user.id and not request.user.has_perm(
        "aa_forum.manage_forum"
    ):
        messages.error(
            request,
            mark_safe(
                _("<h4>Error!</h4><p>You are not allowed to delete this message.</p>")
            ),
        )

        logger.info(
            f"{request.user} was trying to delete message ID {message_id} without "
            f"permission to do so. Redirecting to forum index"
        )

        return redirect(
            "aa_forum:forum_topic",
            category_slug=topic.board.category.slug,
            board_slug=topic.board.slug,
            topic_slug=topic.slug,
        )

    # Let's check if we have more than one message in this topic
    # and the message we want to delete is not the first
    # If so, remove just that message and return to the topic
    if topic.first_message.pk != message.pk and topic.messages.all().count() > 1:
        message.delete()

        messages.success(
            request,
            mark_safe(_("<h4>Success!</h4><p>The message has been deleted.</p>")),
        )

        logger.info(
            f"{request.user} removed message ID {message_id} "
            f'from topic "{topic_subject}"'
        )

        return redirect(
            "aa_forum:forum_topic",
            category_slug=topic.board.category.slug,
            board_slug=topic.board.slug,
            topic_slug=topic.slug,
        )

    # If it is the topics opening post ...
    topic.delete()

    messages.success(
        request,
        mark_safe(
            _(
                "<h4>Success!</h4><p>The message has been deleted.</p>"
                "<p>This was the topics opening post, so the topic has been "
                "deleted as well</p>"
            )
        ),
    )

    logger.info(
        f"{request.user} removed message ID {message_id}. This was the original post, "
        f'so the topic "{topic_subject}" has been removed as well.'
    )

    return redirect(
        "aa_forum:forum_board",
        category_slug=topic.board.category.slug,
        board_slug=topic.board.slug,
    )


@login_required
@permission_required("aa_forum.basic_access")
def mark_all_as_read(request: WSGIRequest) -> HttpResponseRedirect:
    """
    Mark all available topics as read
    :param request:
    :type request:
    :return:
    :rtype:
    """

    has_read_all_messages = LastMessageSeen.objects.filter(
        topic=OuterRef("pk"),
        user=request.user,
        message_time__gte=OuterRef("last_message__time_posted"),
    )

    boards = (
        Board.objects.prefetch_related(
            Prefetch(
                "topics",
                queryset=Topic.objects.select_related(
                    "last_message",
                ).annotate(has_unread_messages=~Exists(has_read_all_messages)),
            )
        )
        .user_has_access(request.user)
        .all()
    )

    if boards.count() > 0:
        for board in boards:
            for topic in board.topics.all():
                LastMessageSeen.objects.update_or_create(
                    topic=topic,
                    user=request.user,
                    defaults={"message_time": topic.last_message.time_posted},
                )

    logger.info(f"{request.user} marked all topics as read")

    return redirect("aa_forum:forum_index")


@login_required
@permission_required("aa_forum.basic_access")
def unread_topics_count(request: WSGIRequest) -> int:
    """
    Get the number of unread messages for the user
    :param request:
    :type request:
    :return:
    :rtype:
    """

    has_read_all_messages = LastMessageSeen.objects.filter(
        topic=OuterRef("pk"),
        user=request.user,
        message_time__gte=OuterRef("last_message__time_posted"),
    )
    unread_topic_pks = Topic.objects.filter(~Exists(has_read_all_messages)).values_list(
        "pk", flat=True
    )

    boards = (
        Board.objects.annotate(
            num_unread_topics=Count(
                "topics", filter=Q(topics__in=unread_topic_pks), distinct=True
            ),
        )
        .user_has_access(request.user)
        .all()
    )

    count_unread_topics = 0

    if boards.count() > 0:
        for board in boards:
            count_unread_topics += board.num_unread_topics

    return count_unread_topics



"""
Tests for our models
"""

# Standard Library
import datetime as dt
from unittest.mock import patch

# Django
from django.contrib.auth.models import Group, User
from django.test import TestCase
from django.urls import reverse
from django.utils.timezone import now

# AA Forum
from aa_forum.models import Board, get_sentinel_user
from aa_forum.tests.utils import (
    create_board,
    create_category,
    create_fake_messages,
    create_fake_user,
    create_last_message_seen,
    create_message,
    create_personal_message,
    create_setting,
    create_topic,
    my_get_setting,
)

MODELS_PATH = "aa_forum.models"


class TestGetSentinelUser(TestCase):
    def test_should_create_user_when_it_does_not_exist(self):
        # when
        user = get_sentinel_user()
        # then
        self.assertEqual(user.username, "deleted")

    def test_should_return_user_when_it_does(self):
        # given
        User.objects.create_user(username="deleted")
        # when
        user = get_sentinel_user()
        # then
        self.assertEqual(user.username, "deleted")


class TestBoard(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        cls.user = create_fake_user(1001, "Bruce Wayne")
        cls.group = Group.objects.create(name="Superhero")

    def setUp(self) -> None:
        self.category = create_category(name="Science")

    def test_model_string_names(self):
        board = create_board(name="Physics", category=self.category)

        self.assertEqual(str(board), "Physics")

    def test_should_update_last_message_when_message_created(self):
        # given
        board = create_board(category=self.category)
        topic = create_topic(board=board)

        # when
        message = create_message(topic=topic, user_created=self.user)

        # then
        board.refresh_from_db()
        self.assertEqual(board.last_message, message)

    def test_should_update_last_message_when_message_created_in_child_board(self):
        # given
        board = create_board(category=self.category)
        child_board = create_board(category=self.category, parent_board=board)
        topic_board = create_topic(board=board)
        topic_child_board = create_topic(board=child_board)
        create_message(topic=topic_board, user_created=self.user)

        # when
        message_child_board = create_message(
            topic=topic_child_board, user_created=self.user
        )

        # then
        board.refresh_from_db()
        self.assertEqual(board.last_message, message_child_board)

    def test_should_return_url(self):
        # given
        board = create_board(category=self.category, name="Physics")

        # when
        url = board.get_absolute_url()

        # then
        self.assertURLEqual(
            url,
            reverse("aa_forum:forum_board", args=["science", "physics"]),
        )

    def test_should_return_board_with_no_groups(self):
        # given
        board = create_board(category=self.category)

        # when
        result = Board.objects.user_has_access(self.user).get(pk=board.pk)

        # then
        self.assertTrue(result)

    def test_should_return_board_for_group_member(self):
        # given
        board = create_board(category=self.category)
        board.groups.add(self.group)

        self.user.groups.add(self.group)

        # when
        result = Board.objects.user_has_access(self.user).get(pk=board.pk)

        # then
        self.assertTrue(result)

    def test_should_return_child_board_for_group_member(self):
        # given
        board = create_board(category=self.category)
        board.groups.add(self.group)

        self.user.groups.add(self.group)

        child_board = create_board(category=self.category, parent_board=board)

        # when
        result = Board.objects.user_has_access(self.user).get(pk=child_board.pk)

        # then
        self.assertTrue(result)

    def test_should_not_return_board_for_non_group_member(self):
        # given
        board = create_board(category=self.category)
        board.groups.add(self.group)

        with self.assertRaises(Board.DoesNotExist):
            Board.objects.user_has_access(self.user).get(pk=board.pk)

    def test_should_not_return_child_board_for_non_group_member(self):
        # given
        board = create_board(category=self.category)
        board.groups.add(self.group)
        board_2 = create_board(category=self.category, parent_board=board)

        with self.assertRaises(Board.DoesNotExist):
            Board.objects.user_has_access(self.user).get(pk=board_2.pk)

    def test_should_generate_new_slug_when_slug_already_exists(self):
        # given
        board = create_board(category=self.category)
        board.slug = "dummy"  # we are faking the same slug here
        board.save()

        # when
        board = create_board(name="Dummy", category=self.category)

        # then
        self.assertEqual(board.slug, "dummy-1")

    def test_should_not_allow_creation_with_hyphen_slugs(self):
        # when
        board = create_board(name="Dummy", category=self.category, slug="-")

        # then
        self.assertEqual(board.slug, "dummy")

    @patch(MODELS_PATH + ".INTERNAL_URL_PREFIX", "-")
    def test_should_create_slug_with_name_matching_internal_url_prefix(self):
        # when
        category = create_board(name="-", category=self.category)

        # then
        self.assertEqual(category.slug, "hyphen")

    def test_should_handle_empty_board_slug(self):
        """
        Handling an empty slug that is returned by Board.save() when the board name is
        just special characters, which are getting stripped by slug generation
        :return:
        """

        # when
        board = create_board(name="@#$%", category=self.category)

        # then
        expected_slug = f"{board.__class__.__name__}-{board.pk}".lower()
        self.assertEqual(board.slug, expected_slug)


class TestCategory(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        cls.user = create_fake_user(1001, "Bruce Wayne")

    def test_model_string_names(self):
        category = create_category(name="Science")

        self.assertEqual(str(category), "Science")

    def test_should_create_new_category_with_slug(self):
        # when
        category = create_category(name="Science")

        # then
        self.assertEqual(category.slug, "science")

    def test_should_keep_existing_slug_when_changing_name(self):
        category = create_category(name="Science")

        # when
        category.name = "Politics"
        category.save()

        # then
        category.refresh_from_db()
        self.assertEqual(category.slug, "science")

    def test_should_generate_new_slug_when_slug_already_exists(self):
        # given
        category = create_category(name="Science")
        category.slug = "dummy"  # we are faking the same slug here
        category.save()

        # when
        category = create_category(name="Dummy")

        # then
        self.assertEqual(category.slug, "dummy-1")

    def test_should_not_allow_creation_with_hyphen_slugs(self):
        # when
        category = create_category(name="Science", slug="-")

        # then
        self.assertEqual(category.slug, "science")

    def test_should_not_allow_to_update_slug_to_hyphen(self):
        # given
        category = create_category(name="Science")

        # when
        category.slug = "-"
        category.save()

        # then
        category.refresh_from_db()
        self.assertEqual(category.slug, "science-1")

    @patch(MODELS_PATH + ".INTERNAL_URL_PREFIX", "-")
    def test_should_create_new_category_with_name_matching_internal_url_prefix(self):
        # when
        category = create_category(name="-")

        # then
        self.assertEqual(category.slug, "hyphen")

    def test_should_handle_empty_category_slug(self):
        """
        Handling an empty slug that is returned by Category.save() when the category
        name is just special characters, which are getting stripped by slug generation
        :return:
        """

        # when
        category = create_category(name="@#$%")

        # then
        expected_slug = f"{category.__class__.__name__}-{category.pk}".lower()
        self.assertEqual(category.slug, expected_slug)


@patch(MODELS_PATH + ".Setting.objects.get_setting", new=my_get_setting)
class TestMessage(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        cls.user = create_fake_user(1001, "Bruce Wayne")

    def setUp(self) -> None:
        category = create_category()

        self.board = create_board(category=category)
        self.child_board = create_board(category=category, parent_board=self.board)
        self.topic = create_topic(board=self.board)
        self.child_topic = create_topic(board=self.child_board)

    def test_model_string_names(self):
        # when
        message = create_message(topic=self.topic, user_created=self.user)

        # then
        self.assertEqual(str(message), str(message.pk))

    def test_should_update_first_and_last_messages_when_saving_1(self):
        # when
        message = create_message(topic=self.topic, user_created=self.user)

        # then
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(self.topic.last_message, message)
        self.assertEqual(self.topic.first_message, message)
        self.assertEqual(self.board.last_message, message)
        self.assertEqual(self.board.first_message, message)

    def test_should_update_first_and_last_messages_when_saving_2(self):
        # given
        message_1 = create_message(topic=self.topic, user_created=self.user)

        # when
        message_2 = create_message(topic=self.topic, user_created=self.user)

        # then
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(self.topic.last_message, message_2)
        self.assertEqual(self.topic.first_message, message_1)
        self.assertEqual(self.board.last_message, message_2)
        self.assertEqual(self.board.first_message, message_1)

    def test_should_update_first_and_last_messages_when_saving_3(self):
        # when
        message = create_message(topic=self.child_topic, user_created=self.user)

        # then
        self.child_topic.refresh_from_db()
        self.child_board.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(self.child_topic.last_message, message)
        self.assertEqual(self.child_topic.first_message, message)
        self.assertEqual(self.child_board.last_message, message)
        self.assertEqual(self.child_board.first_message, message)
        self.assertEqual(self.board.last_message, message)
        self.assertEqual(self.board.first_message, message)

    def test_should_update_first_and_last_messages_when_saving_4(self):
        # given
        message = create_message(topic=self.topic, user_created=self.user)

        # when
        child_message = create_message(topic=self.child_topic, user_created=self.user)

        # then
        self.child_topic.refresh_from_db()
        self.child_board.refresh_from_db()
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(self.child_topic.last_message, child_message)
        self.assertEqual(self.child_topic.first_message, child_message)
        self.assertEqual(self.child_board.last_message, child_message)
        self.assertEqual(self.child_board.first_message, child_message)
        self.assertEqual(self.topic.last_message, message)
        self.assertEqual(self.topic.last_message, message)
        self.assertEqual(self.board.last_message, child_message)
        self.assertEqual(self.board.first_message, child_message)

    def test_should_update_first_and_last_messages_when_saving_5(self):
        # given
        message_1 = create_message(topic=self.topic, user_created=self.user)
        message_2 = create_message(topic=self.topic, user_created=self.user)

        # when
        child_message_1 = create_message(topic=self.child_topic, user_created=self.user)
        child_message_2 = create_message(topic=self.child_topic, user_created=self.user)

        # then
        self.child_topic.refresh_from_db()
        self.child_board.refresh_from_db()
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(self.child_topic.first_message, child_message_1)
        self.assertEqual(self.child_topic.last_message, child_message_2)
        self.assertEqual(self.child_board.first_message, child_message_1)
        self.assertEqual(self.child_board.last_message, child_message_2)
        self.assertEqual(self.topic.first_message, message_1)
        self.assertEqual(self.topic.last_message, message_2)
        self.assertEqual(self.board.first_message, child_message_1)
        self.assertEqual(self.board.last_message, child_message_2)

    def test_should_update_message_references_when_deleting_last_message_1(self):
        # given
        message_1 = create_message(topic=self.topic, user_created=self.user)
        message_2 = create_message(topic=self.topic, user_created=self.user)

        # when
        message_2.delete()

        # then
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(self.topic.last_message, message_1)
        self.assertEqual(self.topic.first_message, message_1)
        self.assertEqual(self.board.last_message, message_1)
        self.assertEqual(self.board.first_message, message_1)

    def test_should_update_message_references_when_deleting_last_message_2(self):
        # given
        message_1 = create_message(topic=self.topic, user_created=self.user)
        message_2 = create_message(topic=self.topic, user_created=self.user)

        # when
        message_1.delete()

        # then
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(self.topic.last_message, message_2)
        self.assertEqual(self.topic.first_message, message_2)
        self.assertEqual(self.board.last_message, message_2)
        self.assertEqual(self.board.first_message, message_2)

    def test_should_update_message_references_when_deleting_last_message_3(self):
        # given
        message = create_message(topic=self.topic, user_created=self.user)
        child_message = create_message(topic=self.child_topic, user_created=self.user)

        # when
        child_message.delete()

        # then
        self.topic.refresh_from_db()
        self.child_topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(self.topic.last_message, message)
        self.assertEqual(self.topic.first_message, message)
        self.assertIsNone(self.child_topic.last_message)
        self.assertIsNone(self.child_topic.first_message)
        self.assertEqual(self.board.last_message, message)
        self.assertEqual(self.board.first_message, message)

    def test_should_update_message_references_when_deleting_last_message_4(self):
        # given
        message = create_message(topic=self.topic, user_created=self.user)
        child_message = create_message(topic=self.child_topic, user_created=self.user)

        # when
        message.delete()

        # then
        self.topic.refresh_from_db()
        self.child_topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertIsNone(self.topic.last_message)
        self.assertIsNone(self.topic.first_message)
        self.assertEqual(self.child_topic.last_message, child_message)
        self.assertEqual(self.child_topic.first_message, child_message)
        self.assertEqual(self.board.last_message, child_message)
        self.assertEqual(self.board.first_message, child_message)

    def test_should_reset_message_references_when_deleting_last_message_1(self):
        # given
        message = create_message(topic=self.topic, user_created=self.user)

        # when
        message.delete()

        # then
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertIsNone(self.topic.last_message)
        self.assertIsNone(self.topic.first_message)
        self.assertIsNone(self.board.last_message)
        self.assertIsNone(self.board.first_message)

    def test_should_return_url_first_page(self):
        # given
        message = create_message(topic=self.topic, user_created=self.user)
        url = message.get_absolute_url()

        # then
        self.assertURLEqual(
            url,
            reverse(
                "aa_forum:forum_topic",
                args=(
                    self.topic.board.category.slug,
                    self.topic.board.slug,
                    self.topic.slug,
                ),
            )
            + f"#message-{message.pk}",
        )

    def test_should_return_url_other_pages(self):
        # given
        create_fake_messages(self.topic, 9)
        message = self.topic.messages.order_by("time_posted").last()
        url = message.get_absolute_url()

        # then
        self.assertURLEqual(
            url,
            reverse(
                "aa_forum:forum_topic",
                args=(
                    self.topic.board.category.slug,
                    self.topic.board.slug,
                    self.topic.slug,
                    2,
                ),
            )
            + f"#message-{message.pk}",
        )


class TestTopic(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        cls.user = create_fake_user(1001, "Bruce Wayne")

    def setUp(self) -> None:
        category = create_category(name="Science")

        self.board = create_board(category=category, name="Physics")
        self.child_board = create_board(category=category, parent_board=self.board)

    def test_model_string_names(self):
        topic = create_topic(subject="Mysteries", board=self.board)

        self.assertEqual(str(topic), "Mysteries")

    def test_should_update_last_message_when_message_is_created(self):
        # given
        topic = create_topic(subject="Mysteries", board=self.board)

        # when
        message = create_message(topic=topic, user_created=self.user)

        # then
        topic.refresh_from_db()
        self.assertEqual(topic.last_message, message)

    def test_should_update_last_message_when_message_is_created_child_board(self):
        # given
        child_topic = create_topic(board=self.child_board)

        # when
        child_message = create_message(topic=child_topic, user_created=self.user)

        # then
        child_topic.refresh_from_db()
        self.assertEqual(child_topic.last_message, child_message)

    def test_should_update_last_message_after_topic_deletion(self):
        # given
        topic_1 = create_topic(board=self.board)
        topic_2 = create_topic(board=self.board)
        my_now = now() - dt.timedelta(hours=1)

        with patch("django.utils.timezone.now", lambda: my_now):
            message_1 = create_message(topic=topic_1, user_created=self.user)

        message_2 = create_message(topic=topic_2, user_created=self.user)

        self.board.refresh_from_db()
        self.assertEqual(self.board.last_message, message_2)

        # when
        topic_2.delete()

        # then
        self.board.refresh_from_db()
        self.assertEqual(self.board.last_message, message_1)

    def test_should_not_update_last_message_after_topic_deletion(self):
        # given
        topic_1 = create_topic(board=self.board)
        topic_2 = create_topic(board=self.board)
        my_now = now() - dt.timedelta(hours=1)

        with patch("django.utils.timezone.now", lambda: my_now):
            create_message(topic=topic_1, user_created=self.user)

        message_2 = create_message(topic=topic_2, user_created=self.user)

        self.board.refresh_from_db()
        self.assertEqual(self.board.last_message, message_2)

        # when
        topic_1.delete()

        # then
        self.board.refresh_from_db()
        self.assertEqual(self.board.last_message, message_2)

    def test_should_return_url(self):
        # given
        topic = create_topic(subject="Mysteries", board=self.board)

        # when
        url = topic.get_absolute_url()

        # then
        self.assertURLEqual(
            url,
            reverse("aa_forum:forum_topic", args=["science", "physics", "mysteries"]),
        )

    def test_should_generate_new_slug_when_slug_already_exists(self):
        # given
        topic = create_topic(board=self.board)
        topic.slug = "dummy"  # we are faking the same slug here
        topic.save()

        # when
        topic = create_topic(subject="Dummy", board=self.board)

        # then
        self.assertEqual(topic.slug, "dummy-1")

    def test_should_not_allow_creation_with_hyphen_slugs(self):
        # when
        topic = create_topic(subject="Dummy", board=self.board, slug="-")

        # then
        self.assertEqual(topic.slug, "dummy")

    @patch(MODELS_PATH + ".INTERNAL_URL_PREFIX", "-")
    def test_should_create_slug_with_name_matching_internal_url_prefix(self):
        # when
        topic = create_topic(subject="-", board=self.board)

        # then
        self.assertEqual(topic.slug, "hyphen")

    def test_should_handle_empty_topic_slug(self):
        """
        Handling an empty slug that is returned by Topic.save() when the topic
        name is just special characters, which are getting stripped by slug generation
        :return:
        """

        # when
        topic = create_topic(subject="@#$%", board=self.board)

        # then
        expected_slug = f"{topic.__class__.__name__}-{topic.pk}".lower()
        self.assertEqual(topic.slug, expected_slug)


class TestPersonalMessage(TestCase):
    def test_str(self):
        # with
        message = create_personal_message(subject="Subject")

        # then
        self.assertEqual(str(message), "Subject")

    @patch(MODELS_PATH + ".INTERNAL_URL_PREFIX", "-")
    def test_can_create_personal_message(self):
        # with
        message = create_personal_message(subject="subject")

        # then
        self.assertEqual(message.subject, "subject")

    def test_should_generate_new_slug_when_slug_already_exists(self):
        # given
        message_1 = create_personal_message()
        message_1.slug = "dummy"  # we are faking the same slug here
        message_1.save()

        # when
        message_2 = create_personal_message(subject="Dummy")

        # then
        self.assertEqual(message_2.slug, "dummy-1")

    def test_should_not_allow_creation_with_hyphen_slugs(self):
        # when
        message = create_personal_message(subject="Dummy", slug="-")

        # then
        self.assertEqual(message.slug, "dummy")

    def test_should_create_slug_with_name_matching_internal_url_prefix(self):
        # when
        message = create_personal_message(subject="-")

        # then
        self.assertEqual(message.slug, "hyphen")


class LastMessageSeen(TestCase):
    def test_str(self):
        # given
        topic = create_topic(subject="Alpha")
        user = create_fake_user(1001, "Bruce Wayne")
        message_time = dt.datetime(2022, 1, 12, 17, 30)
        message = create_last_message_seen(
            topic=topic, user=user, message_time=message_time
        )

        # when/then
        result = str(message)

        # then
        self.assertEqual(result, "Alpha-Bruce_Wayne-2022-01-12 17:30:00")


class TestSetting(TestCase):
    def test_str(self):
        # given
        setting = create_setting(variable="alpha", value=1)

        # when/then
        self.assertEqual(str(setting), "alpha")



"""
Forms
"""

# Third Party
from ckeditor_uploader.widgets import CKEditorUploadingWidget

# Django
from django import forms
from django.contrib.auth.models import Group
from django.forms import ModelForm
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# AA Forum
from aa_forum.models import Board, Category, Message, Topic


def get_mandatory_form_label_text(text):
    """
    Label text for mandatory form fields
    :param text:
    :type text:
    :return:
    :rtype:
    """

    required_text = _("This field is mandatory")
    required_marker = (
        f'<span aria-label="{required_text}" class="form-required-marker">*</span>'
    )

    return mark_safe(
        f'<span class="form-field-required">{text} {required_marker}</span>'
    )


class SpecialModelChoiceIterator(forms.models.ModelChoiceIterator):
    """
    Variant of Django's ModelChoiceIterator to prevent it from always re-fetching the
    given queryset from database.
    """

    def __iter__(self):
        if self.field.empty_label is not None:
            yield ("", self.field.empty_label)

        queryset = self.queryset

        for obj in queryset:
            yield self.choice(obj)


class SpecialModelMultipleChoiceField(forms.ModelMultipleChoiceField):
    """
    Variant of Django's ModelMultipleChoiceField to prevent it from always
    re-fetching the given queryset from database.
    """

    iterator = SpecialModelChoiceIterator

    def _get_queryset(self):
        return self._queryset

    def _set_queryset(self, queryset):
        self._queryset = queryset
        self.widget.choices = self.choices

    queryset = property(_get_queryset, _set_queryset)


class NewTopicForm(forms.Form):
    """
    New topic form
    """

    subject = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Subject")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Subject")}),
    )

    message = forms.CharField(
        widget=CKEditorUploadingWidget(
            config_name="aa_forum",
            attrs={"rows": 10, "cols": 20, "style": "width: 100%;"},
        ),
        required=True,
        label=get_mandatory_form_label_text(_("Message")),
    )


class EditTopicForm(ModelForm):
    """
    Edit category form
    """

    subject = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Subject")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Subject")}),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Topic
        fields = ["subject"]


class NewCategoryForm(ModelForm):
    """
    New category form
    """

    name = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Name")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Category Name")}),
    )
    boards = forms.CharField(
        required=False,
        label=_("Boards"),
        help_text=_(
            "Boards to be created with this category (One board per line). These "
            "boards will have no group restrictions, so you have to add them later "
            "where needed."
        ),
        widget=forms.Textarea(
            attrs={
                "rows": 10,
                "cols": 20,
                "input_type": "textarea",
                "placeholder": _("Boards"),
            }
        ),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Category
        fields = ["name", "boards"]


class EditCategoryForm(ModelForm):
    """
    Edit category form
    """

    name = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Name")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Category Name")}),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Category
        fields = ["name"]


class EditBoardForm(ModelForm):
    """
    Edit board form
    """

    name = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Name")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Board Name")}),
    )
    description = forms.CharField(
        required=False,
        label=_("Description"),
        widget=forms.Textarea(
            attrs={
                "rows": 10,
                "cols": 20,
                "input_type": "textarea",
                "placeholder": _("Board Description (Optional)"),
            }
        ),
    )
    groups = SpecialModelMultipleChoiceField(
        required=False,
        label=_("Group Restrictions"),
        help_text=_(
            "This will restrict access to this board to the selected groups. If no "
            "groups are selected, everyone who can access the forum can also access "
            "this board. (This setting is optional)"
        ),
        queryset=Group.objects.all(),
    )
    discord_webhook = forms.CharField(
        required=False,
        label=_("Discord Webhook (Optional)"),
        help_text=_(
            "Discord Webhook URL for the channel to post about new topics in this "
            "board. (This setting is optional) "
        ),
        max_length=254,
        widget=forms.TextInput(
            attrs={
                "placeholder": "https://discord.com/api/webhooks/xxxxxxxxxxxxxxxxxx/xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
            }
        ),
    )
    use_webhook_for_replies = forms.BooleanField(
        initial=False,
        required=False,
        label=_("Use this Discord Webhook for replies as well?"),
        help_text=_(
            "When activated every reply to any topic in this board will be "
            "posted to the defined Discord channel. (Child boards are excluded) "
            "Chose wisely! (Default: NO)"
        ),
    )

    def __init__(self, *args, **kwargs):
        groups_queryset = kwargs.pop("groups_queryset", None)

        super().__init__(*args, **kwargs)

        if groups_queryset:
            self.fields["groups"].queryset = groups_queryset

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Board
        fields = [
            "name",
            "description",
            "groups",
            "discord_webhook",
            "use_webhook_for_replies",
        ]


class EditMessageForm(ModelForm):
    """
    Edit message form
    """

    message = forms.CharField(
        widget=CKEditorUploadingWidget(
            config_name="aa_forum",
            attrs={"rows": 10, "cols": 20, "style": "width: 100%;"},
        ),
        required=True,
        label=get_mandatory_form_label_text(_("Message")),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Message
        fields = ["message"]



"""
Constants
"""

# Django
from django.utils.text import slugify

# AA Forum
from aa_forum import __version__

github_url = "https://github.com/ppfeufer/aa-forum"
verbose_name = "AA-Forum - A simple forum for Alliance Auth"
verbose_name_slug = slugify(verbose_name, allow_unicode=True)
user_agent = f"{verbose_name_slug} v{__version__} {github_url}"

# Setting keys
SETTING_MESSAGESPERPAGE = "defaultMaxMessages"
SETTING_TOPICSPERPAGE = "defaultMaxTopics"

# All internal URLs need to start with this prefix
# to prevent conflicts with user generated forum URLs
INTERNAL_URL_PREFIX = "-"

# Default sort order for new categories and boards
DEFAULT_CATEGORY_AND_BOARD_SORT_ORDER = 999999

# Search stop words. These words and characters will be removed from the search phrase
SEARCH_STOPWORDS = [
    '"',
    "<",
    ">",
    "(",
    ")",
    "{",
    "}",
    "a",
    "an",
    "are",
    "as",
    "at",
    "be",
    "if",
    "in",
    "into",
    "is",
    "of",
    "off",
    "on",
    "the",
    "what",
    "which",
    "who",
]

# Discord embed settings
DISCORD_EMBED_COLOR_INFO = 0x5BC0DE
DISCORD_EMBED_COLOR_SUCCESS = 0x5CB85C
DISCORD_EMBED_COLOR_WARNING = 0xF0AD4E
DISCORD_EMBED_COLOR_DANGER = 0xD9534F

DISCORD_EMBED_COLOR_MAP = {
    "info": DISCORD_EMBED_COLOR_INFO,
    "success": DISCORD_EMBED_COLOR_SUCCESS,
    "warning": DISCORD_EMBED_COLOR_WARNING,
    "danger": DISCORD_EMBED_COLOR_DANGER,
}

DISCORD_EMBED_MESSAGE_LENGTH = 1000



"""
Text helper
"""

# Standard Library
import re

# Third Party
from bs4 import BeautifulSoup

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from app_utils.urls import site_absolute_url

# AA Forum
from aa_forum import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def verify_image_url(image_url):
    """
    Verify that the passed text is an image URL.

    We're verifying image URLs for inclusion in Slack/Discord Webhook integration,
    which requires a scheme at the beginning (http(s)) and a file extention at the end
    to render correctly. So, a URL which passes verify_url()
    (like example.com/kitten.gif) might not pass this test. If you need to test that
    the URL is both valid AND an image suitable for the Incoming Webhook integration,
    run it through both verify_url() and verify_image_url().
    """

    return re.match("http", image_url) and re.search(
        r"[gif|jpg|jpeg|png|bmp|webp]$", image_url
    )


def get_image_url(text):
    """
    Extract an image url from the passed text. If there are multiple image urls,
    only the first one will be returned.
    """

    soup = BeautifulSoup(text, "html.parser")

    images = soup.findAll("img")

    if images:
        first_image__src = images[0]["src"]

        logger.debug(f"First Image found: {first_image__src}")

        if not first_image__src.startswith(("http://", "https://")):
            logger.debug("Image has no absolute URL, fixing!")

            absolute_site_url = site_absolute_url()
            first_image__src = f"{absolute_site_url}{first_image__src}"

        if verify_image_url(first_image__src):
            logger.debug(f"Image verified: {first_image__src}")

            return first_image__src

    logger.debug("No images found.")

    return None


def string_cleanup(string: str) -> str:
    """
    Clean up a string by removing JS, CSS and Head tags
    :param string:
    :type string:
    """

    re_head = re.compile(r"<\s*head[^>]*>.*?<\s*/\s*head\s*>", re.S | re.I)
    re_script = re.compile(r"<\s*script[^>]*>.*?<\s*/\s*script\s*>", re.S | re.I)
    re_css = re.compile(r"<\s*style[^>]*>.*?<\s*/\s*style\s*>", re.S | re.I)

    # Strip JS
    string = re_head.sub("", string)
    string = re_script.sub("", string)
    string = re_css.sub("", string)

    return string



"""
Discord webhook helper
"""

# Standard Library
import html

# Third Party
from dhooks_lite import Embed, Footer, Image, Webhook

# Django
from django.utils.html import strip_tags

# Alliance Auth (External Libs)
from app_utils.urls import reverse_absolute

# AA Forum
from aa_forum.constants import DISCORD_EMBED_COLOR_MAP, DISCORD_EMBED_MESSAGE_LENGTH
from aa_forum.helper.eve_images import get_character_portrait_from_evecharacter
from aa_forum.helper.text import get_image_url
from aa_forum.models import Board, Message, Topic


def _prepare_message_for_webhook(message: Message) -> str:
    """
    Preparing the message to be sent with a Discord Webhook

    We have to run strip_tags() twice here.
    1. To remove HTML tags from the text we want to send
    2. After html.unescape() in case that unescaped former escaped HTML tags,
    which now need to be removed as well
    :param message:
    :return:
    """
    return strip_tags(
        html.unescape(
            strip_tags(
                message.message[:DISCORD_EMBED_MESSAGE_LENGTH] + "…"
                if len(message.message) > DISCORD_EMBED_MESSAGE_LENGTH
                else message.message
            )
        )
    )


def send_message_to_discord_webhook(
    board: Board, topic: Topic, message: Message, headline: str
):
    """
    Send a message to a Discord Webhook
    :param board:
    :param topic:
    :param message:
    :param headline:
    :return:
    """

    discord_webhook = Webhook(board.discord_webhook)
    message_to_send = _prepare_message_for_webhook(message=message)
    embed_color = DISCORD_EMBED_COLOR_MAP.get("info", None)
    image_url = get_image_url(message.message)
    author_eve_avatar = get_character_portrait_from_evecharacter(
        character=message.user_created.profile.main_character, size=256
    )
    author_eve_name = message.user_created.profile.main_character.character_name

    if message.pk == message.topic.first_message.pk:
        title = topic.subject
        url = reverse_absolute(
            "aa_forum:forum_topic",
            args=[board.category.slug, board.slug, topic.slug],
        )
    else:
        title = f"Re: {topic.subject}"
        url = reverse_absolute(
            "aa_forum:forum_message",
            args=[board.category.slug, board.slug, topic.slug, message.pk],
        )

    embed = Embed(
        description=message_to_send,
        title=title,
        url=url,
        timestamp=message.time_posted,
        color=embed_color,
        footer=Footer(f"Posted by: {author_eve_name}", author_eve_avatar),
        image=Image(image_url) if image_url else None,
        # thumbnail=Thumbnail(author_eve_avatar),
        # author=Author(
        #     author_eve_name,
        #     # url="https://en.wikipedia.org/wiki/Albert_Einstein",
        #     icon_url=author_eve_avatar,
        # ),
        # fields=[
        #     Field("1st Measurement", "Failed"),
        #     Field("2nd Measurement", "Succeeded"),
        # ],
    )

    discord_webhook.execute(
        headline,
        # username=author_eve_name,
        # avatar_url=author_eve_avatar,
        embeds=[embed],
        wait_for_response=True,
    )



"""
Template filter for search views
"""

# Standard Library
import re

# Django
from django.template.defaulttags import register
from django.utils.safestring import mark_safe

# AA Forum
from aa_forum.constants import SEARCH_STOPWORDS


@register.filter
def highlight_search_term(text: str, search_phrase: str) -> str:
    """
    Highlight the search term in search results
    :param text:
    :type text:
    :param search_phrase:
    :type search_phrase:
    :return:
    :rtype:
    """

    querywords = search_phrase.split()
    search_phrase_terms = [
        word for word in querywords if word.lower() not in SEARCH_STOPWORDS
    ]
    highlighted = text

    for search_term in search_phrase_terms:
        highlighted = re.sub(
            "(?i)(%s)" % (re.escape(search_term)), "{«}\\1{»}", highlighted
        )

    return mark_safe(
        highlighted.replace(
            "{«}", '<span class="aa-forum-search-term-highlight">'
        ).replace("{»}", "</span>")
    )
